
#include "global.h"


#include <algorithm>
#include <math.h>


SIGN sign(double dNumber)
{
	if(dNumber > 0)
		return SIGN_POSITIVE;
	else if(dNumber < 0)
		return SIGN_NEGATIVE;
	else
		return SIGN_NEUTRAL;
}


/*!
 * \brief
 * Sort function to use with STL to sort Point2D variables ascending by their abscissa.
 * 
 * \param ptPoint1
 * Variable 1 to compare.
 * 
 * \param ptPoint2
 * Variable 2 to compare.
 * 
 * \returns
 * Returns true if abscissa of variable 1 is smallen than abscissa of variable 2.
 */
bool SortAbscissaAsc(Point2D ptPoint1, Point2D ptPoint2)
{
	return ptPoint1.x < ptPoint2.x;
}

bool SortDirectionAsc(SPOINTWITHDIRECTION ptPoint1, SPOINTWITHDIRECTION ptPoint2)
{
	return ptPoint1.fProjection < ptPoint2.fProjection;
}

bool SortDirectionDesc(SPOINTWITHDIRECTION ptPoint1, SPOINTWITHDIRECTION ptPoint2)
{
	return ptPoint1.fProjection > ptPoint2.fProjection;
}


Point2D min(Point2D ptComp1, Point2D ptComp2)
{
	Point2D ptTemp;

	ptTemp.x = ptComp1.x < ptComp2.x ? ptComp1.x : ptComp2.x;
	ptTemp.y = ptComp1.y < ptComp2.y ? ptComp1.y : ptComp2.y;

	return ptTemp;
}

Point2D max(Point2D ptComp1, Point2D ptComp2)
{
	Point2D ptTemp;

	ptTemp.x = ptComp1.x > ptComp2.x ? ptComp1.x : ptComp2.x;
	ptTemp.y = ptComp1.y > ptComp2.y ? ptComp1.y : ptComp2.y;

	return ptTemp;
}


/*!
 * \brief
 * Retrieves the point with maximum abscissa.
 * 
 * \param pvPoints
 Pointer to a vector containing the points to search.
 * 
 * \param pptResult
 * Pointer to the variable in which to store the result.
 * 
 * \returns
 * Returns an integer value indicating success or failure.
 * 0 = No Points, 1 = Success, 2 = Success but there are other points with the same abscissa value
 */
int MaximumAbscissa(std::vector<Point2D> * pvPoints, Point2D * pptResult)
{
	if(pvPoints->size() == 0)
		return 0;

	bool bDoubles = false;

	Point2D ptTemp = (*pvPoints)[0];
	for(unsigned int i = 1; i < pvPoints->size(); i++){
		if((*pvPoints)[i].x > ptTemp.x){
			ptTemp = (*pvPoints)[i];
			bDoubles = false;
		}
		else if((*pvPoints)[i].x == ptTemp.x)
			bDoubles = true;
	}

	*pptResult = ptTemp;

	if(bDoubles)
		return 2;
	else
		return 1;
}

/*!
 * \brief
 * Retrieves all points with maximum abscissa (not just one as MaximumAbscissaMultiple).
 * 
 * \param pvPoints
 Pointer to a vector containing the points to search.
 * 
 * \param pvResult
 * Pointer to a vector in which to place the result points.
 * 
 * \returns
 * Returns a boolean value indicating success or failure.
 * true = success, false = failure
 */
bool MaximumAbscissaMultiple(std::vector<Point2D> * pvPoints, std::vector<Point2D> * pvResult)
{
	pvResult->clear();

	Point2D ptTemp;
	int iResult = MaximumAbscissa(pvPoints, &ptTemp);
	if(iResult == 0)
		return false;
	else if(iResult == 1){
		pvResult->push_back(ptTemp);
		return true;
	}
	else{
		for(unsigned int i = 0; i < pvPoints->size(); i++){
			if((*pvPoints)[i].x == ptTemp.x)
				pvResult->push_back((*pvPoints)[i]);
			return true;
		}
		return false;
	}
}

/*!
 * \brief
 * Retrieves the point with the maximum distance from origin along a given direction.
 * 
 * \param pvPoints
 * List of points to search.
 * 
 * \param ptDireccion
 * Direction to search.
 * 
 * \param pptResult
 * Pointer to a Point2D in which to store the result.
 * 
 * \returns
 * Integer indicating success or failur.
 * 0 = No Points, 1 = Success, 2 = Found point but there are other points with the same distance.
 */
int MaximumInDirection(std::vector<Point2D> * pvPoints, Point2D ptDireccion, Point2D * pptResult)
{
	if(pvPoints->size() == 0)
		return 0;

	bool bDoubles = false;

	Point2D ptTemp = (*pvPoints)[0];
	float fLength = (*pvPoints)[0] * ptDireccion;
	for(unsigned int i = 1; i < pvPoints->size(); i++){
		if(fLength < (*pvPoints)[i] * ptTemp){
			ptTemp = (*pvPoints)[i];
			fLength = (*pvPoints)[i] * ptTemp;
			bDoubles = false;
		}
		else if(fLength == (*pvPoints)[i] * ptTemp)
			bDoubles = true;
	}

	*pptResult = ptTemp;

	if(bDoubles)
		return 2;
	else
		return 1;
}

/*!
 * \brief
 * Calculates the (axis-aligned) bounding box for given number of points.
 * 
 * \param pvPoints
 * List of points for which to calculate the bounding box.
 * 
 * \param prctResult
 * Pointer to a rectangle structur in which to store the resulting bounding box.
 * 
 * \returns
 * Returs a boolean value indicating success or failure. True meens success.
 */
bool BoundingBox(std::vector<Point2D> * pvPoints, Rect2D * prctResult)
{
	if(pvPoints->size() == 0)
		return false;

	Point2D ptLeftBottom, ptUpperRight;
	ptLeftBottom = ptUpperRight = (*pvPoints)[0];
	for(unsigned int i = 1; i < pvPoints->size(); i++){
		ptLeftBottom = min(ptLeftBottom, (*pvPoints)[i]);
		ptUpperRight = max(ptUpperRight, (*pvPoints)[i]);
	}

	*prctResult = Rect2D(ptLeftBottom, ptUpperRight);
	return true;
}

/*!
 * \brief
 * Sorts a given list of points ascending by abscissa.
 * 
 * \param pvPoints
 * List of points to sort.
 * 
 * \param pvResult
 * Pointer to a list in which to store the points in sorted order.
 * 
 * \remarks
 * Does not sort the original list but stores the result in a new list.
 */
void SortPointsByAbscissaAscending(std::vector<Point2D> * pvPoints, std::vector<Point2D> * pvResult)
{
	pvResult->clear();

	pvResult->insert(pvResult->begin(), pvPoints->begin(), pvPoints->end());
	std::sort(pvResult->begin(), pvResult->end(), SortAbscissaAsc);
}

/*!
 * \brief
 * Sorts a given list of points ascending by there distance from the origin in a given direction.
 * 
 * \param pvPoints
 * List of points to sort.
 * 
 * \param ptDireccion
 * Vector indicating direction in which to sort.
 * 
 * \param pvResult
 * Pointer to a list in which to store the sorted list of points.
 * 
 * \remarks
 * Does not sort the original list but stores the result in a new list.
 */
void SortPointsByDirectionAscending(std::vector<Point2D> * pvPoints, Point2D ptDireccion, std::vector<Point2D> * pvResult)
{
	pvResult->clear();

	std::vector<SPOINTWITHDIRECTION> vTemp;

	SPOINTWITHDIRECTION spwdPointTemp;
	for(unsigned int i = 0; i < pvPoints->size(); i++){
		spwdPointTemp.ptPoint = (*pvPoints)[i];
		spwdPointTemp.fProjection = spwdPointTemp.ptPoint * ptDireccion;
		vTemp.push_back(spwdPointTemp);
	}

	std::sort(vTemp.begin(), vTemp.end(), SortDirectionAsc);

	for(unsigned int i = 0; i < vTemp.size(); i++)
		pvResult->push_back(vTemp[i].ptPoint);
}

/*!
 * \brief
 * For a given rectangle, calculates the points contained in this rectangle.
 * 
 * \param pvPoints
 * List of points to analyse.
 * 
 * \param rcFrame
 * Rectangle to check.
 * 
 * \param pvResult
 * Pointer to a list in which to store the points found to be contained in the rectangle.
 */
void PointsInRect(std::vector<Point2D> * pvPoints, Rect2D rcFrame, std::vector<Point2D> * pvResult)
{
	pvResult->clear();

	for(unsigned int i = 0; i < pvPoints->size(); i++){
		Point2D ptTemp = (*pvPoints)[i];
		if(ptTemp.x >= rcFrame.ptLeftBottom.x && ptTemp.y >= rcFrame.ptLeftBottom.y && ptTemp.x <= rcFrame.ptUpperRight.x && ptTemp.y <= rcFrame.ptUpperRight.y)
			pvResult->push_back(ptTemp);
	}
}

/*!
 * \brief
 * Sorts a list of points angular (counterclockwise) with respect to a given reference point an the point with minimum ordenate.
 * 
 * \param pvPoints
 * List of points to sort.
 * 
 * \param ptRef
 * Reference point.
 * 
 * \param pvResult
 * Pointer to a list to store the result in.
 * 
 * \remarks
 * Does not sort the original list but stores the result in a new list.
 */
void SortPointsAngular(std::vector<Point2D> * pvPoints, Point2D ptRef, std::vector<Point2D> * pvResult)
{
	pvResult->clear();
	if(pvPoints->size() == 0)
		return;

	std::vector<SPOINTWITHDIRECTION> vPointsQ1, vPointsQ2, vPointsQ3, vPointsQ4;

	// Split up the list of points in respect to the quadrant they a located in
	for(unsigned int i = 0; i < pvPoints->size(); i++){
		Point2D ptTemp = (*pvPoints)[i];
		SPOINTWITHDIRECTION spwdTemp;
		spwdTemp.ptPoint = ptTemp;
		spwdTemp.fProjection = abs(ptTemp.x - ptRef.x);

		if(ptTemp.x >= ptRef.x){
			if(ptTemp.y <= ptRef.y)
				vPointsQ1.push_back(spwdTemp);
			else
				vPointsQ2.push_back(spwdTemp);
		}
		else{
			if(ptTemp.y >= ptRef.y)
				vPointsQ3.push_back(spwdTemp);
			else
				vPointsQ4.push_back(spwdTemp);
		}
	}
	
	// Now sort each of the quadrants (without using angular functions) alternating ascending and descending
	std::sort(vPointsQ1.begin(), vPointsQ1.end(), SortDirectionAsc);
	std::sort(vPointsQ2.begin(), vPointsQ2.end(), SortDirectionDesc);
	std::sort(vPointsQ3.begin(), vPointsQ3.end(), SortDirectionAsc);
	std::sort(vPointsQ4.begin(), vPointsQ4.end(), SortDirectionDesc);
	
	// Rejoin the seprate lists
	std::vector<SPOINTWITHDIRECTION> vJoin;
	vJoin.insert(vJoin.end(), vPointsQ1.begin(), vPointsQ1.end());
	vJoin.insert(vJoin.end(), vPointsQ2.begin(), vPointsQ2.end());
	vJoin.insert(vJoin.end(), vPointsQ3.begin(), vPointsQ3.end());
	vJoin.insert(vJoin.end(), vPointsQ4.begin(), vPointsQ4.end());

	// Now find the one with minimum ordenate but use <= instead of < as for points with same minimal ordenate, the last one is the one with minimum abscissa
	unsigned int iIndex = 0;
	float fMinOrd = vJoin[0].ptPoint.y;
	for(unsigned int i = 1; i < vJoin.size(); i++){
		if(vJoin[i].ptPoint.y <= fMinOrd){
			fMinOrd = vJoin[i].ptPoint.y;
			iIndex = i;
		}
	}
	
	// Now generate the shifted and sorted result list
	for(unsigned int i = iIndex; i < vJoin.size(); i++)
		pvResult->push_back(vJoin[i].ptPoint);
	for(unsigned int i = 0; i < iIndex; i++)
		pvResult->push_back(vJoin[i].ptPoint);
}
